// async function* primes(): AsyncGenerator<number> {
//     // 'Promise' simulates the fact that some time is required to get prime number:
//     yield await new Promise(serve => {
//         setTimeout(() => {
//             serve(2);
//         }, 10); // Each 10 ms...
//     });
//     yield await new Promise(serve => {
//         setTimeout(() => {
//             serve(3);
//         }, 10); // Each 10 ms...
//     });
//     yield await new Promise(serve => {
//         setTimeout(() => {
//             serve(5);
//         }, 10); // Each 10 ms...
//     });
//     yield await new Promise(serve => {
//         setTimeout(() => {
//             serve(7);
//         }, 10); // Each 10 ms...
//     });
//     yield await new Promise(serve => {
//         setTimeout(() => {
//             serve(11);
//         }, 10); // Each 10 ms...
//     });
// }
//
// (async () => {
//     for await (const prime of primes())
//         console.info(prime);
// })(); // Lambda is immediately called for test only...

// https://exploringjs.com/nodejs-shell-scripting/ch_web-streams.html
// https://www.loginradius.com/blog/engineering/guest-post/http-streaming-with-nodejs-and-fetch-api/
import * as file_system from "node:fs";
import {Writable} from 'node:stream';

import * as Parser_with_data_loss from 'partial-json-parser';

const UTM = require('utm-latlng');

function Chunks(stream_reader: ReadableStreamDefaultReader) {
    return {
        async* [Symbol.asyncIterator]() {
            let result = await stream_reader.read();
            while (!result.done) {
                yield result.value;
                result = await stream_reader.read();
            }
        },
    };
}

export const WebStream = () => {
    const decompression_stream = new DecompressionStream("gzip");
    fetch("https://bulk.openweathermap.org/sample/city.list.json.gz").then(async (response) => {
        // 'response.body' is a 'ReadableStream' object:
        console.assert(response.body.constructor.name === 'ReadableStream');
        console.assert(response.body.locked === false /*&& response.body.state === 'readable'*/);

        const data_stream: ReadableStream<number> = response.body.pipeThrough(decompression_stream);
        for await (const chunk of Chunks(data_stream.getReader())) {
            console.assert(data_stream.locked); // 'getReader()' locks the stream...
            // Raw data stream:
            console.assert(chunk.constructor.name === 'Uint8Array');
            // console.info(`Chunk of size ${chunk.length}... with raw data: ${chunk}`);
            // Data stream as (incomplete) JSON. Result *CANNOT* be parsed as JSON:
            // console.info((new TextDecoder()).decode(chunk));
            /**
             * Caution, loss of data ('import * as Parser_with_data_loss from 'partial-json-parser';'):
             */
            const cities: Array<City> = Parser_with_data_loss((new TextDecoder()).decode(chunk));
            console.assert(cities.constructor.name === 'Array');
            break; // Stop, first chunk only for test...
        }
    });
}

export const WebStream_ = () => {
    fetch("https://FranckBarbier.com/Enterprise_Architect/Prison_de_Nantes/Prison_de_Nantes.json").then(async (response) => {
        // 'TextDecoderStream' transforms raw data into strings:
        const data_stream = response.body.pipeThrough(new TextDecoderStream());
        for await (const chunk of Chunks(data_stream.getReader())) {
            console.assert(chunk.constructor.name === 'String');
            console.info(`Chunk of size ${chunk.length}... with string data: ${chunk}`);
            // Stop, first chunk only for test:
            break;
        }
    });
}

interface City {
    "id": number,
    "name": string,
    "state": string,
    "country": string,
    "coord"?: {
        "lon": number,
        "lat": number
    }
    // UTM coord. are computed by the transformation stream:
    "UTM"?: { "Easting": number, "Northing": number, "ZoneNumber": number, "ZoneLetter": string }
}

// https://exploringjs.com/nodejs-shell-scripting/ch_web-streams.html#implementing-custom-transformstreams
class Compute_UTM_from_lat_lon implements Transformer<Uint8Array, Array<City>> {
    private static readonly _Precision = 1;
    private static readonly _UTM = new UTM; // Default Ellipsoid is 'WGS 84'
    private readonly _text_decoder: TextDecoder = new TextDecoder();

    start(controller: TransformStreamDefaultController<Array<City>>): void | Promise<void> {
        console.info('Any initialization?');
    }

    transform(chunk: Uint8Array, controller: TransformStreamDefaultController<Array<City>>): void | Promise<void> {
        /**
         * Caution, loss of data ('import * as Parser_with_data_loss from 'partial-json-parser';'):
         */
        const cities: Array<City> = Parser_with_data_loss(this._text_decoder.decode(chunk));
        for (const city of cities)
            if ('coord' in city)
                city.UTM = Compute_UTM_from_lat_lon._UTM.convertLatLngToUtm(city.coord.lat, city.coord.lon, Compute_UTM_from_lat_lon._Precision)
        controller.enqueue(cities);
        controller.terminate(); // Stop, first chunk only for test...
    }

    flush(controller: TransformStreamDefaultController<Array<City>>): void | Promise<void> {
        // Any finalization?
    }
}

export const Transform = () => {
    const decompression_stream = new DecompressionStream("gzip");
    fetch("https://bulk.openweathermap.org/sample/city.list.json.gz").then(async (response) => {
        const data_stream: ReadableStream<any> = response.body.pipeThrough(decompression_stream);

        // const tranformation = new TransformStream({
        //     transform(chunk, controller) {
        //         controller.enqueue(new TextDecoder().decode(chunk));
        //         controller.terminate(); // Stop, first chunk only for test...
        //     },
        // });

        // 'Transformer' object is passed to 'TransformStream' at creation time:
        const transformation = new TransformStream(new Compute_UTM_from_lat_lon);
        const data_stream_: ReadableStream<Array<City>> = data_stream.pipeThrough(transformation);
        const reader: ReadableStreamDefaultReader<Array<City>> = data_stream_.getReader();
        for await (const cities of Chunks(reader))
            for (const city of cities)
                console.info(city);
        reader.releaseLock();

        // const writableStream = new WritableStream({ // (B)
        //     write(chunk) {
        //         console.log('WRITE: ' + JSON.stringify(chunk));
        //     },
        //     close() {
        //         console.log('CLOSE WritableStream');
        //     },
        // });
    });
}

export const Write = async () => {
    const decompression_stream = new DecompressionStream("gzip");
    fetch("https://bulk.openweathermap.org/sample/city.list.json.gz").then(async (response) => {
        const data_stream: ReadableStream<any> = response.body.pipeThrough(decompression_stream);
    //    await data_stream.pipeTo(Writable.toWeb(file_system.createWriteStream("Cities.txt", {encoding: 'utf-8'})));

         const transformedStream = data_stream.pipeThrough(new CompressionStream('gzip'));
        await transformedStream.pipeTo(Writable.toWeb(file_system.createWriteStream("Cities.gz", {encoding: 'utf-8'})));
    });
}
